#! /usr/bin/python

from twisted.trial import unittest
from twisted.internet import reactor
from twisted.internet.defer import Deferred
from twisted.python import util

noisy = False
if noisy:
    from twisted.python.log import startLogging
    import sys
    startLogging(sys.stdout)

import re, time, sys
import signal

from buildbot.slave.commands import SlaveShellCommand

# test slavecommand.py by running the various commands with a fake
# SlaveBuilder object that logs the calls to sendUpdate()

def findDir():
    # the same directory that holds this script
    return util.sibpath(__file__, ".")

class FakeSlaveBuild:
    pass
        
class FakeSlaveBuilder:
    usePTY = True # TODO: this probably won't work on non-posix

    def __init__(self):
        self.updates = []
        self.failure = None
        self.deferred = Deferred()
        self.basedir = findDir()
    def startBuild(self):
        self.build = FakeSlaveBuild()
    def commandComplete(self, dummy):
        if noisy: print "FakeSlaveBuilder.commandComplete"
        self.completed = 1
        self.deferred.callback(0)
    def commandFailed(self, failure):
        if noisy: print "FakeSlaveBuilder.commandFailed", failure
        self.failure = failure
        self.deferred.callback(1)
    def sendUpdate(self, data):
        if noisy: print "FakeSlaveBuilder.sendUpdate", data
        self.updates.append(data)


class SignalMixin:
    sigchldHandler = None
    
    def setUpClass(self):
        # make sure SIGCHLD handler is installed, as it should be on
        # reactor.run(). problem is reactor may not have been run when this
        # test runs.
        if hasattr(reactor, "_handleSigchld") and hasattr(signal, "SIGCHLD"):
            self.sigchldHandler = signal.signal(signal.SIGCHLD,
                                                reactor._handleSigchld)
    
    def tearDownClass(self):
        if self.sigchldHandler:
            signal.signal(signal.SIGCHLD, self.sigchldHandler)


class SlaveCommandTestCase(SignalMixin, unittest.TestCase):

    def setUp(self):
        self.builder = FakeSlaveBuilder()
        d = self.builder.deferred
        d.addCallback(self.callback)
        self.failed = None
        self.results = None

    def callback(self, failed):
        self.failed = failed
        self.results = self.builder.updates

    def doTest(self, commandfactory, args):
        builder = self.builder
        builder.startBuild()
        stepId = None
        cmd = commandfactory(builder, stepId, args)
        d = cmd.start()
        d.addCallbacks(builder.commandComplete, builder.commandFailed)

        timeout = time.time() + 2
        while not (self.results or self.failed) and time.time() < timeout:
            reactor.iterate(0.01)
        if not (self.results or self.failed):
            self.fail("timeout")
        if self.failed:
            print self.builder.failure
        return self.failed

    def getfile(self, which):
        got = ""
        for r in self.results:
            if r.has_key(which):
                got += r[which]
        return got

    def checkStdout(self, expected):
        expected = expected.replace("\n", "\r\n", 1000) # PTY does this
        got = self.getfile('stdout')
        self.assertEquals(got, expected)
    def getrc(self):
        self.failUnless(self.results[-1].has_key('rc'))
        got = self.results[-1]['rc']
        return got
    def checkrc(self, expected):
        got = self.getrc()
        self.assertEquals(got, expected)
        
    def testShell1(self):
        cmd = sys.executable + " emit.py 0"
        args = {'command': cmd, 'workdir': '.', 'timeout': 5}
        failed = self.doTest(SlaveShellCommand, args)
        self.failIf(failed)
        self.checkStdout("this is stdout\n"
                         "this is stderr\n")
        self.checkrc(0)

    def testShell2(self):
        cmd = sys.executable + " emit.py 1"
        args = {'command': cmd, 'workdir': '.', 'timeout': 5}
        failed = self.doTest(SlaveShellCommand, args)
        self.failIf(failed)
        self.checkStdout("this is stdout\n"
                         "this is stderr\n")
        self.checkrc(1)

    def testShell3(self):
        cmd = sys.executable + " emit.py 0"
        args = {'command': cmd, 'workdir': '.',
                'env': {'EMIT_TEST': "envtest"}, 'timeout': 5}
        failed = self.doTest(SlaveShellCommand, args)
        self.failIf(failed)
        self.checkStdout("this is stdout\n"
                         "this is stderr\n"
                         "EMIT_TEST: envtest\n"
                         )
        self.checkrc(0)

    def testShell4(self):
        cmd = sys.executable + " emit.py 0"
        args = {'command': cmd, 'workdir': "subdir", 'timeout': 5}
        failed = self.doTest(SlaveShellCommand, args)
        self.failIf(failed)
        self.checkStdout("this is stdout in subdir\n"
                         "this is stderr\n")
        self.checkrc(0)

    def testShellZ(self):
        args = {'command': "/bin/EndWorldHungerAndMakePigsFly",
                'workdir': '.', 'timeout': 5}
        failed = self.doTest(SlaveShellCommand, args)
        self.failIf(failed)
        self.failUnless(self.getrc() != 0)
        got = self.getfile('stdout')
        self.failUnless(re.search(r'no such file',
                                  got,
                                  re.I))

    # todo: interrupt(), kill process

    # todo: twisted-specific command tests
