#! /usr/bin/python

import sys, os, signal, shutil, time
from email.Utils import mktime_tz, parsedate_tz

from twisted.trial import unittest
dr = unittest.deferredResult
from twisted.internet import defer, reactor
#defer.Deferred.debug = True

from twisted.python import log
#log.startLogging(sys.stderr)

from buildbot import master, interfaces
from buildbot.slave import bot
from buildbot.status import builder
from buildbot.status.builder import SUCCESS
from buildbot.process import step, base
from buildbot.changes import changes

# Most of these tests (all but SourceStamp) depend upon having a set of
# repositories from which we can perform checkouts. These repositories are
# provided in a small tarball available from the buildbot web site, named
# buildbot-test-vc-1.tar.gz . Unpack this tarball into any convenient
# directory 'foo' and then run the tests with BUILDBOT_TEST_VC=path/to/foo


# CVS requires a local file repository. Providing remote access is beyond
# the feasible abilities of this test program (needs pserver or ssh).

# SVN requires a local file repository (and is irritatingly large: the
# database files for 1kB source tree consume over 5MB). To provide remote
# access over HTTP requires an apache server with DAV support and mod_svn,
# way beyond what we can test from here. Note that is is possible that the
# SVN repository format is not portable across platforms. If you have
# problems running this test but Subversion otherwise works normally for
# you, please file a bug so we can improve the test framework.

# Arch and Darcs both allow remote (read-only) operation with any web
# server. We test both local file access and HTTP access (by spawning a
# small web server to provide access to the repository files while the test
# is running).


RepositoryPath = None
if os.environ.get("BUILDBOT_TEST_VC"):
    p = os.path.join(os.path.expanduser(os.environ.get("BUILDBOT_TEST_VC")),
                     "buildbot-test-vc-1")
    if os.path.exists(os.path.join(p, "README")):
        RepositoryPath = p
        # some of them take a file: URI
        RepositoryPath_URI = "file://" + RepositoryPath
        # many VC systems can access the repository via HTTP
        Repository_HTTP = "http://localhost:%d"

have_arch, have_darcs, have_svn, have_cvs = False, False, False, False
for p in os.environ['PATH'].split(":"):
    if os.path.exists(os.path.join(p, 'tla')):
        have_arch = True
    if os.path.exists(os.path.join(p, 'darcs')):
        have_darcs = True
    if os.path.exists(os.path.join(p, 'svn')):
        # we need svn to be compiled with the ra_local access module
        from twisted.internet import utils
        v = dr(utils.getProcessOutput('svn', ["--version"], env=os.environ))
        if "handles 'file' schema" in v:
            have_svn = True
        else:
            log.msg(("%s found but it does not support 'file:' schema, " +
                     "skipping svn tests") % os.path.join(p, "svn"))
    if os.path.exists(os.path.join(p, 'cvs')):
        have_cvs = True

if not have_svn:
    log.msg("could not find usable 'svn', skipping Subversion tests")
if not have_arch:
    log.msg("could not find 'tla' on $PATH, skipping Arch tests")
if not have_darcs:
    log.msg("could not find 'darcs' on $PATH, skipping Darcs tests")
if not have_cvs:
    log.msg("could not find 'cvs' on $PATH, skipping CVS tests")

config_vc = """
from buildbot.process import factory, step

def s(klass, **kwargs):
    return (klass, kwargs)

f1 = factory.BuildFactory([
    %s,
    ])
c = {}
c['bots'] = [['bot1', 'sekrit']]
c['sources'] = []
c['builders'] = [{'name': 'vc', 'slavename': 'bot1',
                  'builddir': 'vc-dir', 'factory': f1}]
c['slavePortnum'] = 0
BuildmasterConfig = c
"""

p0_diff = r"""
Index: good.c
===================================================================
RCS file: /usr/home/warner/stuff/Projects/BuildBot/fakerep/sample/good.c,v
retrieving revision 1.1
diff -u -r1.1 good.c
--- good.c	10 Nov 2002 10:37:35 -0000	1.1
+++ good.c	31 Aug 2004 00:42:16 -0000
@@ -1,8 +1,10 @@
 #include <stdio.h>
 
+/* this is a hello-world program */
+
 int
 main(int argc, const char *argv[])
 {
-    printf("Hello world\n");
+    printf("Hello, world.\n");
     return 0;
 }
"""

class FakeBuilder:
    name = "fake"
    def getSlaveCommandVersion(self, command, oldversion=None):
        return "1.10"

class SignalMixin:
    sigchldHandler = None
    
    def setUpClass(self):
        # make sure SIGCHLD handler is installed, as it should be on
        # reactor.run(). problem is reactor may not have been run when this
        # test runs.
        if hasattr(reactor, "_handleSigchld") and hasattr(signal, "SIGCHLD"):
            self.sigchldHandler = signal.signal(signal.SIGCHLD,
                                                reactor._handleSigchld)
    
    def tearDownClass(self):
        if self.sigchldHandler:
            signal.signal(signal.SIGCHLD, self.sigchldHandler)

class SetupMixin:
    master = None
    slave = None
    serveHTTP = None
    httpServer = None
    httpPort = None

    def setUp(self):
        if self.serveHTTP:
            # launch an HTTP server to serve the repository files
            from twisted.web import static, server
            from twisted.internet import reactor
            root = static.File(self.serveHTTP)
            self.httpServer = reactor.listenTCP(0, server.Site(root))
            self.httpPort = self.httpServer.getHost().port
        shutil.rmtree("basedir", ignore_errors=1)
        os.mkdir("basedir")
        self.master = master.BuildMaster("basedir")
        self.slavebase = os.path.abspath("slavebase")
        shutil.rmtree(self.slavebase, ignore_errors=1)
        os.mkdir("slavebase")

    def connectSlave(self):
        port = self.master.slavePort._port.getHost().port
        slave = bot.BuildSlave("localhost", port, "bot1", "sekrit",
                               self.slavebase, keepalive=0, usePTY=1)
        self.slave = slave
        slave.startService()
        d = self.master.botmaster.waitUntilBuilderAttached("vc")
        dr(d)

    def loadConfig(self, config):
        # reloading the config file causes a new 'listDirs' command to be
        # sent to the slave. To synchronize on this properly, it is easiest
        # to stop and restart the slave.
        if self.slave:
            d = self.master.botmaster.waitUntilBuilderDetached("vc")
            dr(defer.maybeDeferred(self.slave.stopService))
            dr(d)
        self.master.loadConfig(config)
        self.connectSlave()

    def tearDown(self):
        if self.slave:
            d = self.master.botmaster.waitUntilBuilderDetached("vc")
            dr(defer.maybeDeferred(self.slave.stopService))
            dr(d)
        if self.master:
            dr(defer.maybeDeferred(self.master.stopService))
        if self.httpServer:
            dr(self.httpServer.stopListening())
            from twisted.protocols import http
            http._logDateTimeStop() # shut down the internal timer. DUMB!

    def doBuild(self):
        c = interfaces.IControl(self.master)
        
        bc = c.getBuilder("vc").forceBuild(None, "test_vc forced build")
        b = dr(bc.getStatus().waitUntilFinished())
        r = b.getResults()
        if r != SUCCESS:
            assert b.isFinished()
            print
            print
            print "Build did not succeed:", r, b.getText()
            for s in b.getSteps():
                for l in s.getLogs():
                    print "--- START step %s / log %s ---" % (s.getName(),
                                                              l.getName())
                    print l.getTextWithHeaders()
                    print "--- STOP ---"
                    print
            self.fail("build did not succeed")

    def touch(self, d, f):
        open(os.path.join(d,f),"w").close()
    def shouldExist(self, d, f):
        self.failUnless(os.path.exists(os.path.join(d, f)))
    def shouldNotExist(self, d, f):
        self.failIf(os.path.exists(os.path.join(d, f)))

    def do_vc(self, vctype, args, metadir=None, testRetry=True):
        m = self.master
        vcdir = os.path.join(self.slavebase, "vc-dir", "source")
        workdir = os.path.join(self.slavebase, "vc-dir", "build")
        # woo double-substitution
        s = "s(%s, timeout=20, workdir='build', mode='%%s'" % (vctype,)
        for k,v in args.items():
            s += ", %s='%s'" % (k, v)
        s += ")"
        config = config_vc % s

        m.loadConfig(config % 'clobber')
        m.readConfig = True
        m.startService()

        self.connectSlave()

        self.doBuild() # initial checkout
        self.shouldExist(workdir, "Makefile")
        if metadir:
            self.shouldExist(workdir, metadir)

        self.touch(workdir, "newfile")
        self.shouldExist(workdir, "newfile")
        self.doBuild() # rebuild clobbers workdir

        self.shouldNotExist(workdir, "newfile")

        self.loadConfig(config % 'update')
        self.doBuild() # rebuild with update
        if metadir:
            self.shouldExist(workdir, metadir)

        self.touch(workdir, "newfile")
        self.doBuild() # update rebuild leaves new files
        self.shouldExist(workdir, "newfile")

        if testRetry:
            # certain local changes will prevent an update from working. The
            # most common is to replace a file with a directory, or vice
            # versa. The slave code should spot the failure and do a
            # clobber/retry.
            os.unlink(os.path.join(workdir, "good.c"))
            os.mkdir(os.path.join(workdir, "good.c"))
            self.touch(os.path.join(workdir, "good.c"), "foo")

            self.doBuild() # update, but must clobber to handle the error
            self.shouldNotExist(workdir, "newfile")

        self.loadConfig(config % 'copy')
        self.doBuild() # copy rebuild clobbers new files
        if metadir:
            self.shouldExist(workdir, metadir)
        self.shouldNotExist(workdir, "newfile")
        self.touch(workdir, "newfile")
        self.touch(vcdir, "newvcfile")

        self.doBuild() # copy rebuild clobbers new files
        if metadir:
            self.shouldExist(workdir, metadir)
        self.shouldNotExist(workdir, "newfile")
        self.shouldExist(vcdir, "newvcfile")
        self.shouldExist(workdir, "newvcfile")
        self.touch(workdir, "newfile")

        if metadir:
            self.loadConfig(config % 'export')

            self.doBuild() #"export rebuild clobbers new files
            if metadir:
                self.shouldNotExist(workdir, metadir)
            self.shouldNotExist(workdir, "newfile")
            self.touch(workdir, "newfile")

            self.doBuild() # export rebuild clobbers new files
            if metadir:
                self.shouldNotExist(workdir, metadir)
            self.shouldNotExist(workdir, "newfile")


class VC(SignalMixin, SetupMixin, unittest.TestCase):

    def testCVS(self):
        if not have_cvs:
            raise unittest.SkipTest("CVS is not installed")
        self.do_vc("step.CVS", {
            'cvsroot': RepositoryPath + "/CVS-Repository",
            'cvsmodule': "sample"},
                   metadir="CVS")

    def testSVN(self):
        if not have_svn:
            raise unittest.SkipTest("No usable Subversion was found")
        self.do_vc("step.SVN", {
            'svnurl': RepositoryPath_URI + "/SVN-Repository/sample"},
                   metadir=".svn")

    def testDarcs(self):
        if not have_darcs:
            raise unittest.SkipTest("Darcs is not installed")
        self.do_vc("step.Darcs", {
            'repourl': RepositoryPath_URI + "/Darcs-Repository",
            },
                   testRetry=False)
        # Darcs has a metadir="_darcs", but it does not have an 'export'
        # mode

        # TODO: testRetry has the same problem with Darcs as it does for
        # Arch

    def testArch(self):
        if not have_arch:
            raise unittest.SkipTest("Arch (tla) is not installed")
        self.do_vc("step.Arch", {
            'url': RepositoryPath + "/Arch-Repository",
            'version': "testvc--mainline--1",
            },
                   testRetry=False)
        # Arch has a metadir="{arch}", but it does not have an 'export' mode.

        # the current testRetry=True logic doesn't have the desired effect:
        # "update" is a no-op because arch knows that the repository hasn't
        # changed. Other VC systems will re-checkout missing files on
        # update, arch just leaves the tree untouched. TODO: come up with
        # some better test logic, probably involving a copy of the
        # repository that has a few changes checked in.

        # tell tla to get rid of the leftover archive this test leaves in the
        # user's 'tla archives' listing. The name of this archive is provided
        # by the repository tarball, so the following command must use the
        # same name. We could use archive= to set it explicitly, but if you
        # change it from the default, then 'tla update' won't work.
        os.system("tla register-archive -d test@buildbot.sf.net--testvc")

class VC_HTTP(SignalMixin, SetupMixin, unittest.TestCase):
    serveHTTP = RepositoryPath

    def testDarcsHTTP(self):
        if not have_darcs:
            raise unittest.SkipTest("Darcs is not installed")
        self.do_vc("step.Darcs", {
            'repourl': Repository_HTTP % self.httpPort + "/Darcs-Repository",
            }, testRetry=False)
        # Darcs has a metadir="_darcs", but it does not have an 'export' mode

    def testArchHTTP(self):
        if not have_arch:
            raise unittest.SkipTest("Arch (tla) is not installed")
        self.do_vc("step.Arch", {
            'url': Repository_HTTP % self.httpPort + "/Arch-Repository",
            'version': "testvc--mainline--1",
            }, testRetry=False)
        os.system("tla register-archive -d test@buildbot.sf.net--testvc")

class Patch(SignalMixin, SetupMixin, unittest.TestCase):

    def doPatch(self, vc, revision, **kwargs):
        m = self.master
        config = config_vc % "(step.Dummy)"
        m.loadConfig(config)
        m.readConfig = True
        m.startService()
        self.connectSlave()
        remote = m.botmaster.builders["vc"].remote
        build = base.Build()
        build.builder = FakeBuilder()
        c = vc(mode="clobber", workdir="build", build=build, **kwargs)
        c.step_status = builder.BuildStepStatus(None)
        build.setSourceStamp(revision, (0, p0_diff))
        d = defer.maybeDeferred(c.startStep, remote)
        r = dr(d)
        if r != SUCCESS:
            print c.cmd.log.getTextWithHeaders()
            self.fail("vc+patch command failed, rc=%s" % c.cmd.rc)
        # make sure the file actually got patched
        f = open(os.path.join(self.slavebase, "vc-dir", "build", "good.c"),
                 "r")
        data = f.read()
        self.failUnless("this is a hello-world program" in data)
        self.failUnless("Hello, world.\\n" in data)

        # make sure that a rebuild does not use the leftover patched workdir
        build.setSourceStamp(None, None)
        c = vc(mode="update", workdir="build", build=build, **kwargs)
        c.step_status = builder.BuildStepStatus(None)
        d = defer.maybeDeferred(c.startStep, remote)
        r = dr(d)
        if r != SUCCESS:
            print c.cmd.log.getTextWithHeaders()
            self.fail("vc+patch-then-update command failed, rc=%s" % c.cmd.rc)
        # make sure the file is back to its original
        f = open(os.path.join(self.slavebase, "vc-dir", "build", "good.c"),
                 "r")
        data = f.read()
        self.failUnless("Hello world\\n" in data) # no comma


    def testPatchCVS(self):
        if not have_cvs:
            raise unittest.SkipTest("CVS is not installed")
        self.doPatch(vc=step.CVS,
                     cvsroot= RepositoryPath + "/CVS-Repository",
                     cvsmodule="sample",
                     revision="today")

    def testPatchSVN(self):
        if not have_svn:
            raise unittest.SkipTest("No usable Subversion was found")
        self.doPatch(vc=step.SVN,
                     svnurl= RepositoryPath_URI + "/SVN-Repository/sample",
                     revision="HEAD")


class SourceStamp(unittest.TestCase):
    def addChange(self, b, when=None, revision=None):
        if when:
            when = mktime_tz(parsedate_tz(when))
        c = changes.Change("fred", [], "", when=when, revision=revision)
        b.changes.append(c)

    def testCVS1(self):
        b = base.Build()
        b.treeStableTimer = 100
        s = step.CVS(cvsroot=None, cvsmodule=None, workdir=None, build=b)
        self.failUnlessEqual(s.computeSourceRevision(b.allChanges()), None)

    def testCVS2(self):
        b = base.Build()
        b.treeStableTimer = 100
        self.addChange(b, "Wed, 08 Sep 2004 09:00:00 -0700")
        self.addChange(b, "Wed, 08 Sep 2004 09:01:00 -0700")
        self.addChange(b, "Wed, 08 Sep 2004 09:02:00 -0700")
        s = step.CVS(cvsroot=None, cvsmodule=None, workdir=None, build=b)
        self.failUnlessEqual(s.computeSourceRevision(b.allChanges()),
                             "Wed, 08 Sep 2004 16:02:50 -0000")

    def testCVS3(self):
        b = base.Build()
        b.treeStableTimer = 100
        self.addChange(b, "Wed, 08 Sep 2004 09:00:00 -0700")
        self.addChange(b, "Wed, 08 Sep 2004 09:01:00 -0700")
        self.addChange(b, "Wed, 08 Sep 2004 09:02:00 -0700")
        s = step.CVS(cvsroot=None, cvsmodule=None, workdir=None, build=b,
                     checkoutDelay=10)
        self.failUnlessEqual(s.computeSourceRevision(b.allChanges()),
                             "Wed, 08 Sep 2004 16:02:10 -0000")

    def testSVN1(self):
        b = base.Build()
        b.treeStableTimer = 100
        s = step.SVN(svnurl=None, workdir=None, build=b)
        self.failUnlessEqual(s.computeSourceRevision(b.allChanges()), None)

    def testSVN2(self):
        b = base.Build()
        b.treeStableTimer = 100
        self.addChange(b, revision=4)
        self.addChange(b, revision=10)
        self.addChange(b, revision=67)
        s = step.SVN(svnurl=None, workdir=None, build=b)
        self.failUnlessEqual(s.computeSourceRevision(b.allChanges()), 67)


if not RepositoryPath:
    VC.skip = "provide BUILDBOT_TEST_VC=path/to/repositories to run this"
    VC_HTTP.skip = "provide BUILDBOT_TEST_VC=path/to/repositories to run this"
    Patch.skip = "provide BUILDBOT_TEST_VC=path/to/repositories to run this"
